import re
from contextlib import contextmanager
from . import utils, sh
from . import bootstrap
from . import ra
from . import corosync
from . import log
from . import xmlutil
from . import constants
from . import sbd
from .service_manager import ServiceManager

logger = log.setup_logger(__name__)
logger_utils = log.LoggerUtils(logger)


class Error(ValueError):
    def __init__(self, msg: str):
        super().__init__(msg)


class ClusterFSManager(object):
    """
    Class to manage cluster filesystem (OCFS2 or GFS2)
    and configure related resources
    """
    def __init__(self, context):
        """
        Init function
        """
        self.ocfs2_devices = context.ocfs2_devices
        self.gfs2_devices = context.gfs2_devices
        self.stage = context.stage
        self.use_cluster_lvm2 = context.use_cluster_lvm2
        self.mount_point = context.mount_point
        self.use_stage = context.stage in ("ocfs2", "gfs2")
        self.yes_to_all = context.yes_to_all
        self.exist_ra_id_list = []
        self.vg_id = None
        self.group_id = None
        # device that finally mounted
        self.target_device = None
        self.type = None
        self.devices = []

        self._verify_options()

        if self.ocfs2_devices:
            self.type = "OCFS2"
            self.devices = self.ocfs2_devices
        elif self.gfs2_devices:
            self.type = "GFS2"
            self.devices = self.gfs2_devices
        if self.type:
            prefix = self.type.lower()
            self.fstype = prefix

            # Consider such issue might not be fixed in the local pacemaker version
            # https://github.com/ClusterLabs/pacemaker/pull/3766
            # DLM RA's id shouldn avoid using 'gfs' as prefix
            self.DLM_RA_ID = "dlm-controld-ra"
            self.FS_RA_ID = f"{prefix}-clusterfs"
            self.LVMLOCKD_RA_ID = f"{prefix}-lvmlockd"
            self.LVMACTIVATE_RA_ID = f"{prefix}-lvmactivate"
            self.GROUP_ID = f"{prefix}-group"
            self.CLONE_ID = f"{prefix}-clone"
            self.VG_ID = f"{prefix}-vg"
            self.LV_ID = f"{prefix}-lv"

    def _verify_packages(self, fs_type=None, use_cluster_lvm2=None):
        """
        Find if missing required package
        """
        _type = fs_type or self.type
        _use_cluster_lvm2 = use_cluster_lvm2 or self.use_cluster_lvm2
        cluster_nodes = utils.list_cluster_nodes() if self.use_stage else []

        package_requirements = {
            "GFS2": ["gfs2-utils"],
            "OCFS2": ["ocfs2-tools"]
        }
        required_packages = package_requirements.get(_type, [])
        if _use_cluster_lvm2:
            required_packages.append("lvm2-lockd")

        def check_packages(node=None):
            for pkg in required_packages:
                if not utils.package_is_installed(pkg, node):
                    node_info = f" on {node}" if node else ""
                    raise Error(f"Missing required package for configuring {_type}{node_info}: {pkg}")

        for node in cluster_nodes:
            check_packages(node)
        else:
            check_packages()

    def _verify_options(self):
        """
        Verify options related with OCFS2 and GFS2
        """
        if self.gfs2_devices and self.ocfs2_devices:
            raise Error("Can't use -g and -o options together")
        if self.stage == "ocfs2" and not self.ocfs2_devices:
            raise Error("ocfs2 stage require -o option")
        if self.stage == "gfs2" and not self.gfs2_devices:
            raise Error("gfs2 stage require -g option")
        if self.use_cluster_lvm2:
            if not self.ocfs2_devices and not self.gfs2_devices:
                raise Error("-C option only valid together with -o or -g option")
        else:
            if len(self.ocfs2_devices) > 1:
                raise Error("Without Cluster LVM2 (-C option), -o option only support one device")
            elif len(self.gfs2_devices) > 1:
                raise Error("Without Cluster LVM2 (-C option), -g option only support one device")
        if self.mount_point and utils.has_mount_point_used(self.mount_point):
            raise Error(f"Mount point {self.mount_point} already mounted")

    def _verify_devices(self):
        """
        Verify OCFS2/GFS2 devices
        """
        for dev in self.devices:
            if not utils.is_block_device(dev):
                raise Error(f"{dev} doesn't look like a block device")
            if utils.is_dev_used_for_lvm(dev) and self.use_cluster_lvm2:
                raise Error(f"{dev} is a Logical Volume, cannot be used with the -C option")
            if utils.has_disk_mounted(dev):
                raise Error(f"{dev} is already mounted")

    def _check_if_already_configured(self):
        """
        Check if OCFS2/GFS2 related resource already configured
        """
        if not self.use_stage:
            return
        out = sh.cluster_shell().get_stdout_or_raise_error("crm configure show")
        if f"fstype={self.fstype}" in out:
            logger.warning("Already configured %s related resources", self.type)
            raise utils.TerminateSubCommand

    def _pre_verify(self):
        """
        Verify before configuring on init process
        """
        self._verify_packages()
        self._check_if_already_configured()
        self._verify_devices()

    @property
    def error_hints_for_stage(self):
        hints = f"""
The cluster service has already been initialized, but the prerequisites are missing
to configure {self.type}. Please fix it and use the stage procedure to configure {self.type} separately,
e.g. crm cluster init {self.type.lower()} -o <device>
        """
        return "" if self.use_stage else hints

    def _check_device_with_sbd_device(self):
        """
        Raise error when OCFS2/GFS2 device is the same with sbd device
        """
        if ServiceManager().service_is_enabled(constants.SBD_SERVICE):
            sbd_device_list = sbd.SBDUtils.get_sbd_device_from_config()
            for dev in self.devices:
                if dev in sbd_device_list:
                    msg = f"{dev} cannot be the same with SBD device" + self.error_hints_for_stage
                    raise Error(msg)

    def _confirm_to_overwrite_device(self):
        """
        Confirm to overwrit OCFS2/GFS2 device on interactive mode
        """
        for dev in self.devices:
            msg = ""
            if utils.has_dev_partitioned(dev):
                msg = f"Found a partition table in {dev}"
            else:
                fs_type = utils.get_dev_fs_type(dev)
                if fs_type:
                    msg = f"{dev} contains a {fs_type} file system"
            if msg and not bootstrap.confirm(f"{msg} - overwrite?"):
                raise utils.TerminateSubCommand

        shell = sh.cluster_shell()
        for dev in self.devices:
            shell.get_stdout_or_raise_error(f"wipefs -a {dev}")

    def init_verify(self):
        """
        Verify after cluster running on init node
        """
        if not utils.has_stonith_running():
            msg = f"{self.type} requires stonith device configured and running." + self.error_hints_for_stage
            raise Error(msg)

        self._check_device_with_sbd_device()
        self._confirm_to_overwrite_device()

    def _gen_ra_scripts(self, ra_type: str, kv: dict) -> tuple[str, str]:
        """
        Generate ra scripts
        Return id and scripts
        """
        if ra_type not in ra.CONFIGURE_RA_TEMPLATE_DICT:
            raise Error(f"Unsupported RA type: {ra_type}")
        config_scripts = ""
        kv["id"] = utils.gen_unused_id(self.exist_ra_id_list, kv["id"])
        config_scripts = ra.CONFIGURE_RA_TEMPLATE_DICT[ra_type].format(**kv)
        return kv["id"], config_scripts

    def _mkfs(self):
        """
        Creating OCFS2/GFS2 filesystem for the target device
        """
        cluster_name = corosync.get_value('totem.cluster_name')
        mkfs_cmd = ""
        if self.type == "OCFS2":
            # TODO now -N value is fixed to 8, need to be configurable in the future if needed
            mkfs_cmd = f"mkfs.ocfs2 --cluster-stack pcmk --cluster-name {cluster_name} -N 8 -x {self.target_device}"
        elif self.type == "GFS2":
            # TODO make sure the lock table name is real unique in the future if needed
            lock_table_name = f"{cluster_name}:FS_{utils.randomword(12)}"
            # TODO now -j value is fixed to 8, need to be configurable in the future if needed
            mkfs_cmd = f"mkfs.gfs2 -t {lock_table_name} -p lock_dlm -j 8 {self.target_device} -O"
        logger.debug("mkfs command: %s", mkfs_cmd)
        with logger_utils.status_long(f"Creating {self.fstype} filesystem on {self.target_device}"):
            sh.cluster_shell().get_stdout_or_raise_error(mkfs_cmd)

    @contextmanager
    def _vg_change(self):
        """
        vgchange process using contextmanager
        """
        shell = sh.cluster_shell()
        shell.get_stdout_or_raise_error(f"vgchange -ay {self.vg_id}")
        try:
            yield
        finally:
            shell.get_stdout_or_raise_error(f"vgchange -an {self.vg_id}")

    def _create_lv(self):
        """
        Create PV, VG, LV and return LV path
        """
        disks_string = ' '.join(self.devices)
        shell = sh.cluster_shell()

        # Create PV
        with logger_utils.status_long(f"Creating PV for {disks_string}"):
            shell.get_stdout_or_raise_error(f"pvcreate {disks_string} -y")

        # Create VG
        self.vg_id = utils.gen_unused_id(utils.get_all_vg_name(), self.VG_ID)
        with logger_utils.status_long(f"Creating VG {self.vg_id}"):
            shell.get_stdout_or_raise_error(f"vgcreate --shared {self.vg_id} {disks_string} -y")

        # Create LV
        with logger_utils.status_long(f"Creating LV {self.LV_ID} on VG {self.vg_id}"):
            pe_number = utils.get_pe_number(self.vg_id)
            shell.get_stdout_or_raise_error(f"lvcreate -l {pe_number} {self.vg_id} -n {self.LV_ID} -y")
 
        return f"/dev/{self.vg_id}/{self.LV_ID}"

    def _gen_group_and_clone_scripts(self, ra_list: list) -> str:
        """
        Generate group and clone scripts
        """
        # Group
        group_kv = {"id":self.GROUP_ID, "ra_string":' '.join(ra_list)}
        self.group_id, group_scripts = self._gen_ra_scripts("GROUP", group_kv)
        # Clone
        clone_kv = {"id":self.CLONE_ID, "group_id":self.group_id}
        _, clone_scripts = self._gen_ra_scripts("CLONE", clone_kv)
        return group_scripts + clone_scripts

    def _gen_fs_scripts(self):
        """
        Generate Filesystem scripts
        """
        fs_kv = {
                "id": self.FS_RA_ID,
                "mnt_point": self.mount_point,
                "fs_type": self.fstype,
                "device": self.target_device
                }
        return self._gen_ra_scripts("Filesystem", fs_kv)

    def _load_append_and_wait(self, scripts, res_id, msg, need_append=True):
        """
        Load scripts, append to exist group and wait resource started
        """
        bootstrap.crm_configure_load("update", scripts)
        if need_append:
            utils.append_res_to_group(self.group_id, res_id)
        bootstrap.wait_for_resource(msg, res_id)

    def _config_dlm(self):
        """
        Configure DLM resource
        """
        config_scripts = ""
        dlm_id, dlm_scripts = self._gen_ra_scripts("DLM", {"id":self.DLM_RA_ID})
        group_clone_scripts = self._gen_group_and_clone_scripts([dlm_id])
        config_scripts = dlm_scripts + group_clone_scripts
        msg = f"Wait for DLM ({dlm_id}) start"
        self._load_append_and_wait(config_scripts, dlm_id, msg, need_append=False)

    def _config_lvmlockd(self):
        """
        Configure LVMLockd resource
        """
        _id, _scripts = self._gen_ra_scripts("LVMLockd", {"id":self.LVMLOCKD_RA_ID})
        msg = f"Wait for LVMLockd ({_id}) start"
        self._load_append_and_wait(_scripts, _id, msg)

    def _config_lvmactivate(self):
        """
        Configure LVMActivate resource
        """
        _id, _scripts = self._gen_ra_scripts("LVMActivate", {"id": self.LVMACTIVATE_RA_ID, "vgname": self.vg_id})
        msg = f"Wait for LVMActivate ({_id}) start"
        self._load_append_and_wait(_scripts, _id, msg)

    def _config_fs(self):
        """
        Configure Filesystem resource
        """
        utils.mkdirp(self.mount_point)
        _id, _scripts = self._gen_fs_scripts()
        msg = f"Wait for Filesystem ({_id}) start"
        self._load_append_and_wait(_scripts, _id, msg)

    def _configure_resource_stack(self):
        self._config_dlm()

        if self.use_cluster_lvm2:
            # Configure dlm + lvmlockd + lvm-activate + Filesystem
            self._config_lvmlockd()
            self.target_device = self._create_lv()
            with self._vg_change():
                self._mkfs()
            self._config_lvmactivate()
        else:
            # Configure dlm + Filesystem
            self.target_device = self.devices[0]
            self._mkfs()

        self._config_fs()

        logger.info("%s device %s mounted on %s", self.type, self.target_device, self.mount_point)

    def init(self):
        """
        OCFS2/GFS2 configure process on init node
        """
        logger.info("Configuring %s", self.type)

        self.init_verify()

        self.exist_ra_id_list = utils.all_exist_id()

        no_quorum_policy_value = utils.get_property("no-quorum-policy")
        if not no_quorum_policy_value or no_quorum_policy_value != "freeze":
            utils.set_property("no-quorum-policy", "freeze")

        self._configure_resource_stack()

    def _find_target_on_join(self, peer) -> dict:
        """
        Find device name from OCF2/GFS2 Filesystem param on peer node
        Return (cluster_fs_type, device)
        """
        fstype_pattern = r'fstype=(ocfs2|gfs2)'
        device_pattern = r'device="([^"]+)"'
        pattern = rf'{fstype_pattern}.*{device_pattern}|{device_pattern}.*{fstype_pattern}'
        out = sh.cluster_shell().get_stdout_or_raise_error("crm configure show", peer)
        res = re.search(pattern, out)
        if res:
            cluster_fs_type = res.group(1) or res.group(4)
            device = res.group(2) or res.group(3)
            return {"cluster_fs_type": cluster_fs_type, "device": device}
        return None

    def join(self, peer):
        """
        Called on join process, to verify OCFS2/GFS2 environment
        """
        target_dict = self._find_target_on_join(peer)
        if not target_dict:
            return
        cluster_fs_type, device = target_dict["cluster_fs_type"], target_dict["device"]
        with logger_utils.status_long(f"Verify {cluster_fs_type.upper()} environment on {device}"):
            use_cluster_lvm2 = xmlutil.CrmMonXmlParser(peer).is_resource_configured(constants.LVMLOCKD_RA)
            self._verify_packages(cluster_fs_type.upper(), use_cluster_lvm2)
            if utils.is_dev_a_plain_raw_disk_or_partition(device, peer):
                utils.compare_uuid_with_peer_dev([device], peer)

    @classmethod
    def pre_verify(cls, ctx):
        inst = cls(ctx)
        inst._pre_verify()



# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
#
# unit tests for parse.py

import copy
import unittest
import pytest
from unittest import mock
from crmsh import corosync


def test_query_status_exception():
    with pytest.raises(ValueError) as err:
        corosync.query_status("test")
    assert str(err.value) == "Wrong type \"test\" to query status"


@mock.patch('crmsh.sh.cluster_shell')
@mock.patch('crmsh.corosync.query_ring_status')
def test_query_status(mock_ring_status, mock_cluster_shell):
    mock_cluster_shell_inst = mock.Mock()
    mock_cluster_shell.return_value = mock_cluster_shell_inst
    mock_cluster_shell_inst.get_stdout_or_raise_error.return_value = "data"
    corosync.query_status("ring")
    mock_ring_status.assert_called_once_with()


@mock.patch('crmsh.corosync.is_qdevice_configured')
def test_query_qdevice_status_exception(mock_configured):
    mock_configured.return_value = False
    with pytest.raises(ValueError) as err:
        corosync.query_qdevice_status()
    assert str(err.value) == "QDevice/QNetd not configured!"
    mock_configured.assert_called_once_with()


@mock.patch('crmsh.sh.ClusterShell.get_stdout_or_raise_error')
@mock.patch('crmsh.corosync.is_qdevice_configured')
def test_query_qdevice_status(mock_configured, mock_run):
    mock_configured.return_value = True
    corosync.query_qdevice_status()
    mock_run.assert_called_once_with("corosync-qdevice-tool -sv")


@mock.patch('crmsh.sh.cluster_shell')
@mock.patch("crmsh.corosync.query_ring_status")
def test_query_status_ring(mock_ring_status, mock_cluster_shell):
    mock_cluster_shell_inst = mock.Mock()
    mock_cluster_shell.return_value = mock_cluster_shell_inst
    mock_cluster_shell_inst.get_stdout_or_raise_error.return_value = "data"
    corosync.query_status("ring")
    mock_ring_status.assert_called_once_with()


@mock.patch('crmsh.sh.cluster_shell')
@mock.patch("crmsh.corosync.query_quorum_status")
def test_query_status_quorum(mock_quorum_status, mock_cluster_shell):
    mock_cluster_shell_inst = mock.Mock()
    mock_cluster_shell.return_value = mock_cluster_shell_inst
    mock_cluster_shell_inst.get_stdout_or_raise_error.return_value = "data"
    corosync.query_status("quorum")
    mock_quorum_status.assert_called_once_with()


@mock.patch('crmsh.sh.cluster_shell')
@mock.patch("crmsh.corosync.query_qnetd_status")
def test_query_status_qnetd(mock_qnetd_status, mock_cluster_shell):
    mock_cluster_shell_inst = mock.Mock()
    mock_cluster_shell.return_value = mock_cluster_shell_inst
    mock_cluster_shell_inst.get_stdout_or_raise_error.return_value = "data"
    corosync.query_status("qnetd")
    mock_qnetd_status.assert_called_once_with()


def test_query_status_except():
    with pytest.raises(ValueError) as err:
        corosync.query_status("xxx")
    assert str(err.value) == "Wrong type \"xxx\" to query status"


@mock.patch("crmsh.sh.ShellUtils.get_stdout_stderr")
def test_query_ring_status_except(mock_run):
    mock_run.return_value = (1, None, "error")
    with pytest.raises(ValueError) as err:
        corosync.query_ring_status()
    assert str(err.value) == "error"
    mock_run.assert_called_once_with("corosync-cfgtool -s")


@mock.patch("crmsh.sh.ShellUtils.get_stdout_stderr")
def test_query_ring_status(mock_run):
    mock_run.return_value = (0, "data", None)
    corosync.query_ring_status()
    mock_run.assert_called_once_with("corosync-cfgtool -s")


@mock.patch("crmsh.sh.ShellUtils.get_stdout_stderr")
def test_query_quorum_status_except(mock_run):
    mock_run.return_value = (1, None, "error")
    with pytest.raises(ValueError) as err:
        corosync.query_quorum_status()
    assert str(err.value) == "error"
    mock_run.assert_called_once_with("corosync-quorumtool -s")


@mock.patch("crmsh.sh.ShellUtils.get_stdout_stderr")
def test_query_quorum_status(mock_run):
    mock_run.return_value = (0, "data", None)
    corosync.query_quorum_status()
    mock_run.assert_called_once_with("corosync-quorumtool -s")


@mock.patch("crmsh.sh.ShellUtils.get_stdout_stderr")
def test_query_quorum_status_no_quorum(mock_run):
    mock_run.return_value = (2, "no quorum", None)
    corosync.query_quorum_status()
    mock_run.assert_called_once_with("corosync-quorumtool -s")


@mock.patch("crmsh.corosync.is_qdevice_configured")
def test_query_qnetd_status_no_qdevice(mock_qdevice_configured):
    mock_qdevice_configured.return_value = False
    with pytest.raises(ValueError) as err:
        corosync.query_qnetd_status()
    assert str(err.value) == "QDevice/QNetd not configured!"
    mock_qdevice_configured.assert_called_once_with()


@mock.patch("crmsh.corosync.get_value")
@mock.patch("crmsh.corosync.is_qdevice_configured")
def test_query_qnetd_status_no_cluster_name(mock_qdevice_configured, mock_get_value):
    mock_qdevice_configured.return_value = True
    mock_get_value.return_value = None
    with pytest.raises(ValueError) as err:
        corosync.query_qnetd_status()
    assert str(err.value) == "cluster_name not configured!"
    mock_qdevice_configured.assert_called_once_with()
    mock_get_value.assert_called_once_with("totem.cluster_name")


@mock.patch("crmsh.corosync.get_value")
@mock.patch("crmsh.corosync.is_qdevice_configured")
def test_query_qnetd_status_no_host(mock_qdevice_configured, mock_get_value):
    mock_qdevice_configured.return_value = True
    mock_get_value.side_effect = ["hacluster", None]
    with pytest.raises(ValueError) as err:
        corosync.query_qnetd_status()
    assert str(err.value) == "host for qnetd not configured!"
    mock_qdevice_configured.assert_called_once_with()
    mock_get_value.assert_has_calls([
        mock.call("totem.cluster_name"),
        mock.call("quorum.device.net.host")
        ])


@mock.patch('crmsh.utils.user_pair_for_ssh')
@mock.patch("crmsh.parallax.parallax_call")
@mock.patch("crmsh.corosync.get_value")
@mock.patch("crmsh.corosync.is_qdevice_configured")
def test_query_qnetd_status_copy_id_failed(mock_qdevice_configured,
        mock_get_value, mock_parallax_call, mock_user_pair_for_ssh):
    mock_user_pair_for_ssh.return_value = "alice", "root"
    mock_parallax_call.side_effect = ValueError("Failed on 10.10.10.123: foo")
    mock_qdevice_configured.return_value = True
    mock_get_value.side_effect = ["hacluster", "10.10.10.123"]
    with pytest.raises(ValueError) as err:
        corosync.query_qnetd_status()
    assert err.value.args[0] == "Failed on 10.10.10.123: foo"
    mock_qdevice_configured.assert_called_once_with()
    mock_get_value.assert_has_calls([
        mock.call("totem.cluster_name"),
        mock.call("quorum.device.net.host")
        ])


@mock.patch('crmsh.utils.user_pair_for_ssh')
@mock.patch("crmsh.parallax.parallax_call")
@mock.patch("crmsh.corosync.get_value")
@mock.patch("crmsh.corosync.is_qdevice_configured")
def test_query_qnetd_status_copy(mock_qdevice_configured, mock_get_value,
        mock_parallax_call, mock_user_pair_for_ssh):
    mock_user_pair_for_ssh.return_value = "alice", "root"
    mock_qdevice_configured.return_value = True
    mock_get_value.side_effect = ["hacluster", "10.10.10.123"]
    mock_parallax_call.return_value = [("node1", (0, "data", None)), ]

    corosync.query_qnetd_status()

    mock_qdevice_configured.assert_called_once_with()
    mock_get_value.assert_has_calls([
        mock.call("totem.cluster_name"),
        mock.call("quorum.device.net.host")
        ])
    mock_parallax_call.assert_called_once_with(["10.10.10.123"], "corosync-qnetd-tool -lv -c hacluster")


@mock.patch('crmsh.corosync.get_corosync_value_dict')
def test_token_and_consensus_timeout(mock_get_dict):
    mock_get_dict.return_value = {"token": 10, "consensus": 12}
    assert corosync.token_and_consensus_timeout() == 22


@mock.patch('crmsh.corosync.get_corosync_value')
def test_get_corosync_value_dict(mock_get_value):
    mock_get_value.side_effect = ["10000", None]
    res = corosync.get_corosync_value_dict()
    assert res == {"token": 10, "consensus": 12}


@mock.patch('crmsh.corosync.get_value')
@mock.patch('crmsh.sh.ClusterShell.get_stdout_or_raise_error')
def test_get_corosync_value_raise(mock_run, mock_get_value):
    mock_run.side_effect = ValueError
    mock_get_value.return_value = None
    assert corosync.get_corosync_value("xxx") is None
    mock_run.assert_called_once_with("corosync-cmapctl runtime.config.xxx")
    mock_get_value.assert_called_once_with("xxx")


@mock.patch('crmsh.sh.ClusterShell.get_stdout_or_raise_error')
def test_get_corosync_value(mock_run):
    mock_run.return_value = "totem.token = 10000"
    assert corosync.get_corosync_value("totem.token") == "10000"
    mock_run.assert_called_once_with("corosync-cmapctl runtime.config.totem.token")


class TestConfigParserSet(unittest.TestCase):
    def setUp(self) -> None:
        self.inst = corosync.ConfParser(config_data='')

    def test_set_scalar_should_ignore_index(self):
        self.inst._raw_set('scalar.scalar', 'foo', 0)
        self.assertDictEqual({'scalar': {'scalar': 'foo'}}, self.inst._dom)
        self.inst._raw_set('scalar.scalar', 'bar', 1)
        self.assertDictEqual({'scalar': {'scalar': 'bar'}}, self.inst._dom)

    def test_set_vector(self):
        self.inst._dom = {'vector': []}
        self.inst._raw_set('vector.scalar', 'foo', 0)
        self.assertDictEqual({'vector': [{'scalar': 'foo'}]}, self.inst._dom)
        self.inst._raw_set('vector.scalar', 'bar', 1)
        self.assertDictEqual({'vector': [{'scalar': 'foo'}, {'scalar': 'bar'}]}, self.inst._dom)
        with self.assertRaises(IndexError):
            self.inst._raw_set('vector.scalar', 'bar', 3)

    def test_set_predefined_vector(self):
        self.inst._raw_set('totem.interface.foo', 0, 0)
        self.assertDictEqual({'totem': {'interface': {'foo': 0}}}, self.inst._dom)
        self.inst._raw_set('totem.interface.foo', 0, 1)
        self.assertDictEqual({'totem': {'interface': [{'foo': 0}, {'foo': 0}]}}, self.inst._dom)
        with self.assertRaises(IndexError):
            self.inst._raw_set('totem.interface.foo', 0, 3)


class TestLinkLoadOptions(unittest.TestCase):
    def test_load_int(self):
        link = corosync.Link()
        link.load_options({'mcastport': '1234'})
        self.assertEqual(1234, link.mcastport)
        self.assertIsNone(link.knet_link_priority)
        self.assertIsNone(link.knet_transport)

    def test_load_int_invalid(self):
        link = corosync.Link()
        with self.assertRaises(ValueError):
            link.load_options({'mcastport': 'sctp'})

    def test_load_str(self):
        link = corosync.Link()
        link.load_options({'knet_transport': 'sctp'})
        self.assertIsNone(link.mcastport)
        self.assertIsNone(link.knet_link_priority)
        self.assertEqual('sctp', link.knet_transport)


class TestLinkManagerGetTotemTransport(unittest.TestCase):
    def test_get_value_from_config(self):
        lm = corosync.LinkManager({'totem': {'transport': 'udpu'}})
        self.assertEqual('udpu', lm.totem_transport())

    def test_get_value_from_default(self):
        lm = corosync.LinkManager({'totem': {'foo': 'bar'}})
        self.assertEqual('knet', lm.totem_transport())


class TestLinkManagerShowLinks(unittest.TestCase):
    def test_non_knet(self):
        lm = corosync.LinkManager({'totem': {'transport': 'udpu'}})
        with self.assertRaises(AssertionError):
            lm.links()

    def test_link_without_options(self):
        lm = corosync.LinkManager({
            'totem': {
                'interface': [{
                    'linknumber': '0',
                    'knet_link_priority': '1',
                }, {
                    'linknumber': '2',
                    'knet_link_priority': '10',
                    'knet_transport': 'sctp',
                }]
            },
            'nodelist': {
                'node': [{
                    'nodeid': '1',
                    'name': 'node1',
                    'ring0_addr': '192.0.2.1',
                    'ring3_addr': '192.0.2.101',
                    'ring2_addr': '192.0.2.201',
                }, {
                    'nodeid': '3',
                    'name': 'node3',
                    'ring0_addr': '192.0.2.3',
                    'ring3_addr': '192.0.2.103',
                    'ring2_addr': '192.0.2.203',
                }, {
                    'nodeid': '2',
                    'name': 'node2',
                    'ring0_addr': '192.0.2.2',
                    'ring3_addr': '192.0.2.102',
                    'ring2_addr': '192.0.2.202',
                }]
            }
        })
        links = lm.links()
        self.assertEqual(8, len(links))
        self.assertTrue(all(links[i] is None for i in [1, 4, 5, 6, 7]))
        self.assertEqual(3, links[3].linknumber)
        self.assertEqual(3, len(links[3].nodes))
        self.assertEqual(1, links[3].nodes[0].nodeid)
        self.assertEqual(2, links[3].nodes[1].nodeid)
        self.assertEqual(3, links[3].nodes[2].nodeid)
        self.assertEqual(1, links[0].knet_link_priority)
        self.assertIsNone(links[3].knet_link_priority)
        self.assertEqual(10, links[2].knet_link_priority)
        self.assertEqual('sctp', links[2].knet_transport)

    def test_only_one_node(self):
        lm = corosync.LinkManager({
            'nodelist': {
                'node': [{
                    'nodeid': '1',
                    'name': 'node1',
                    'ring0_addr': '192.0.2.1',
                }]
            }
        })
        links = lm.links()
        self.assertEqual(8, len(links))
        self.assertTrue(all(links[i] is None for i in range(1, 8)))
        self.assertEqual(0, links[0].linknumber)
        self.assertEqual(1, len(links[0].nodes))
        self.assertEqual(1, links[0].nodes[0].nodeid)
        self.assertIsNone(links[0].knet_link_priority)


class TestLinkManagerUpdateLink(unittest.TestCase):
    ORIGINAL = {
        'totem': {
            'interface': [{
                'linknumber': '0',
                'knet_link_priority': '1',
            }, {
                'linknumber': '2',
                'knet_link_priority': '10',
                'knet_transport': 'sctp',
            }]
        },
        'nodelist': {
            'node': [{
                'nodeid': '1',
                'name': 'node1',
                'ring0_addr': '192.0.2.1',
                'ring1_addr': '192.0.2.101',
                'ring2_addr': '192.0.2.201',
            }, {
                'nodeid': '3',
                'name': 'node3',
                'ring0_addr': '192.0.2.3',
                'ring1_addr': '192.0.2.103',
                'ring2_addr': '192.0.2.203',
            }, {
                'nodeid': '2',
                'name': 'node2',
                'ring0_addr': '192.0.2.2',
                'ring1_addr': '192.0.2.102',
                'ring2_addr': '192.0.2.202',
            }]
        }
    }

    def setUp(self):
        self.lm = corosync.LinkManager(copy.deepcopy(self.ORIGINAL))

    def test_update_and_add_new_option(self):
        dom = self.lm.update_link(0, {'knet_transport': 'sctp', 'knet_link_priority': '2'})
        self.assertEqual(2, len(dom['totem']['interface']))
        self.assertDictEqual({
            'linknumber': '0',
            'knet_link_priority': '2',
            'knet_transport': 'sctp'
        }, dom['totem']['interface'][0])
        self.assertDictEqual(self.ORIGINAL['totem']['interface'][1], dom['totem']['interface'][1])
        self.assertDictEqual(self.ORIGINAL['nodelist'], dom['nodelist'])

    def test_add_new_interface_section(self):
        dom = self.lm.update_link(1, {'knet_link_priority': '2'})
        self.assertEqual(3, len(dom['totem']['interface']))
        self.assertDictEqual({
            'linknumber': '1',
            'knet_link_priority': '2',
        }, dom['totem']['interface'][2])
        self.assertDictEqual(self.ORIGINAL['totem']['interface'][0], dom['totem']['interface'][0])
        self.assertDictEqual(self.ORIGINAL['totem']['interface'][1], dom['totem']['interface'][1])
        self.assertDictEqual(self.ORIGINAL['nodelist'], dom['nodelist'])

    def test_remove_option(self):
        dom = self.lm.update_link(2, {'knet_transport': None})
        self.assertEqual(2, len(dom['totem']['interface']))
        self.assertEqual('2', dom['totem']['interface'][1]['linknumber'])
        self.assertNotIn('knet_transport', dom['totem']['interface'][1])
        self.assertDictEqual(self.ORIGINAL['nodelist'], dom['nodelist'])

    def test_remove_interface_section(self):
        dom = self.lm.update_link(0, {'knet_link_priority': None})
        self.assertEqual(1, len(dom['totem']['interface']))
        self.assertEqual('2', dom['totem']['interface'][0]['linknumber'])
        self.assertDictEqual(self.ORIGINAL['nodelist'], dom['nodelist'])

    def test_add_non_unsupported_option(self):
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'knet_link_priority': '2', 'foo': 'bar'})
        self.assertDictEqual({
            'linknumber': '0',
            'knet_link_priority': '1',
        }, self.lm._config['totem']['interface'][0])
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'linknumber': '1'})
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'nodes': [{'foo': 'bar'}]})

    def test_remove_non_unsupported_option(self):
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'knet_link_priority': '2', 'foo': None})
        self.assertDictEqual({
            'linknumber': '0',
            'knet_link_priority': '1',
        }, self.lm._config['totem']['interface'][0])
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'linknumber': None})
        with self.assertRaises(ValueError):
            self.lm.update_link(0, {'nodes': None})

    def test_update_non_existing_link(self):
        with self.assertRaises(ValueError):
            self.lm.update_link(3, dict())


class TestLinkManagerUpdateNodeAddr(unittest.TestCase):
    ORIGINAL = {
        'nodelist': {
            'node': [{
                'nodeid': '1',
                'name': 'node1',
                'ring0_addr': '192.0.2.1',
                'ring1_addr': '192.0.2.101',
                'ring2_addr': '192.0.2.201',
            }, {
                'nodeid': '3',
                'name': 'node3',
                'ring0_addr': '192.0.2.3',
                'ring1_addr': '192.0.2.103',
                'ring2_addr': '192.0.2.203',
            }, {
                'nodeid': '2',
                'name': 'node2',
                'ring0_addr': '192.0.2.2',
                'ring1_addr': '192.0.2.102',
                'ring2_addr': '192.0.2.202',
            }]
        }
    }

    def setUp(self):
        self.lm = corosync.LinkManager(copy.deepcopy(self.ORIGINAL))

    def test_update_addr(self):
        self.lm.update_node_addr(
            1,
            {
                1: "192.0.2.65",
                2: "192.0.2.66",
                3: "192.0.2.67",
            }
        )
        self.assertEqual('1', self.lm._config['nodelist']['node'][0]['nodeid'])
        self.assertEqual('3', self.lm._config['nodelist']['node'][1]['nodeid'])
        self.assertEqual('2', self.lm._config['nodelist']['node'][2]['nodeid'])
        self.assertEqual('192.0.2.65', self.lm._config['nodelist']['node'][0]['ring1_addr'])
        self.assertEqual('192.0.2.67', self.lm._config['nodelist']['node'][1]['ring1_addr'])
        self.assertEqual('192.0.2.66', self.lm._config['nodelist']['node'][2]['ring1_addr'])

    def test_update_unchanged_addr(self):
        self.lm.update_node_addr(
            1,
            {
                1: "192.0.2.101",
                2: "192.0.2.102",
                3: "192.0.2.103",
            }
        )
        self.assertListEqual(self.ORIGINAL['nodelist']['node'], self.lm._config['nodelist']['node'])

    def test_update_duplicated_addr(self):
        with self.assertRaises(corosync.LinkManager.DuplicatedNodeAddressException):
            self.lm.update_node_addr(1, {1: "192.0.2.1"})
        with self.assertRaises(corosync.LinkManager.DuplicatedNodeAddressException):
            self.lm.update_node_addr(1, {1: "192.0.2.2"})
        with self.assertRaises(corosync.LinkManager.DuplicatedNodeAddressException):
            self.lm.update_node_addr(1, {1: "192.0.2.102"})
        with self.assertRaises(corosync.LinkManager.DuplicatedNodeAddressException):
            self.lm.update_node_addr(
                1,
                {
                    1: "192.0.2.65",
                    2: "192.0.2.66",
                    3: "192.0.2.65",
                }
            )
        with self.assertRaises(corosync.LinkManager.DuplicatedNodeAddressException):
            self.lm.update_node_addr(
                1,
                {
                    1: "fd00:a0::1",
                    2: "fd00:a0::2",
                    3: "fd00:00a0:0000:0000:0000:0000:0000:0001",
                }
            )

    def test_update_unknown_node(self):
        with self.assertRaises(ValueError):
            self.lm.update_node_addr(
                1,
                {
                    1: "192.0.2.65",
                    4: "192.0.2.66",
                }
            )
        self.assertEqual('1', self.lm._config['nodelist']['node'][0]['nodeid'])
        self.assertEqual('192.0.2.101', self.lm._config['nodelist']['node'][0]['ring1_addr'])

    def test_update_unknown_link(self):
        with self.assertRaises(ValueError):
            self.lm.update_node_addr(
                3,
                {
                    1: "192.0.2.65",
                    2: "192.0.2.66",
                    3: "192.0.2.67",
                }
            )
        self.assertDictEqual(self.ORIGINAL, self.lm._config)


class TestLinkManagerRemoveLink(unittest.TestCase):
    ORIGINAL = {
        'totem': {
            'interface': [{
                'linknumber': '0',
                'knet_link_priority': '1',
            }, {
                'linknumber': '2',
                'knet_link_priority': '10',
                'knet_transport': 'sctp',
            }]
        },
        'nodelist': {
            'node': [{
                'nodeid': '1',
                'name': 'node1',
                'ring0_addr': '192.0.2.1',
                'ring1_addr': '192.0.2.101',
                'ring2_addr': '192.0.2.201',
            }, {
                'nodeid': '3',
                'name': 'node3',
                'ring0_addr': '192.0.2.3',
                'ring1_addr': '192.0.2.103',
                'ring2_addr': '192.0.2.203',
            }, {
                'nodeid': '2',
                'name': 'node2',
                'ring0_addr': '192.0.2.2',
                'ring1_addr': '192.0.2.102',
                'ring2_addr': '192.0.2.202',
            }]
        }
    }

    def setUp(self):
        self.lm = corosync.LinkManager(copy.deepcopy(self.ORIGINAL))

    def test_remove(self):
        self.lm.remove_link(1)
        self.assertEqual(2, len(self.lm._config['totem']['interface']))
        self.assertEqual('0', self.lm._config['totem']['interface'][0]['linknumber'])
        self.assertEqual('2', self.lm._config['totem']['interface'][1]['linknumber'])
        self.assertEqual(3, len(self.lm._config['nodelist']['node']))
        self.assertNotIn('ring1_addr', self.lm._config['nodelist']['node'][0])
        self.assertNotIn('ring1_addr', self.lm._config['nodelist']['node'][1])
        self.assertNotIn('ring1_addr', self.lm._config['nodelist']['node'][2])
        self.assertEqual('192.0.2.201', self.lm._config['nodelist']['node'][0]['ring2_addr'])
        self.assertEqual('192.0.2.203', self.lm._config['nodelist']['node'][1]['ring2_addr'])
        self.assertEqual('192.0.2.202', self.lm._config['nodelist']['node'][2]['ring2_addr'])

    def test_remove_unknown_link(self):
        with self.assertRaises(ValueError):
            self.lm.remove_link(3)

    def test_remove_last_link(self):
        self.lm.remove_link(1)
        self.lm.remove_link(0)
        self.assertEqual(1, len(self.lm._config['totem']['interface']))
        self.assertIn('ring2_addr', self.lm._config['nodelist']['node'][0])
        self.assertIn('ring2_addr', self.lm._config['nodelist']['node'][1])
        self.assertIn('ring2_addr', self.lm._config['nodelist']['node'][2])
        with self.assertRaises(ValueError):
            self.lm.remove_link(2)


@mock.patch('crmsh.corosync.LinkManager.update_link')
@mock.patch('crmsh.corosync.LinkManager._LinkManager__upsert_node_addr_impl')
@mock.patch('crmsh.corosync.LinkManager.links')
class TestLinkManagerAddLink(unittest.TestCase):
    def test_unspecified_node(self, mock_links, mock_upsert_node, mock_update_link):
        mock_links.return_value = [None] * 8
        mock_links.return_value[0] = corosync.Link(0, [
            corosync.LinkNode(1, 'node1', '192.0.2.101'),
            corosync.LinkNode(2, 'node2', '192.0.2.102'),
        ])
        lm = corosync.LinkManager(dict())
        with self.assertRaises(corosync.LinkManager.MissingNodesException):
            lm.add_link({1: '192.0.2.201'}, dict())
        mock_upsert_node.assert_not_called()
        mock_update_link.assert_not_called()

    def test_unknown_node(self, mock_links, mock_upsert_node, mock_update_link):
        mock_links.return_value = [None] * 8
        mock_links.return_value[0] = corosync.Link(0, [
            corosync.LinkNode(1, 'node1', '192.0.2.101'),
        ])
        mock_upsert_node.side_effect = ValueError()
        lm = corosync.LinkManager(dict())
        with self.assertRaises(ValueError):
            lm.add_link({1: '192.0.2.201', 2: '192.0.2.202'}, dict())
        mock_upsert_node.assert_called_once()
        mock_update_link.assert_not_called()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
import os
import crm_script
data = crm_script.get_input()
health_report = data[1]

print("Processing collected information...")

CORE_PACKAGES = ['corosync', 'pacemaker', 'resource-agents']

warnings = []
errors = []


def warn(fmt, *args):
    warnings.append(fmt % args)


def error(fmt, *args):
    errors.append(fmt % args)


# sort {package: {version: [host]}}
rpm_versions = {}

LOW_UPTIME = 60.0
HIGH_LOAD = 1.0

for node, info in health_report.items():
    if node != info['system']['hostname']:
        error("Hostname mismatch: %s is not %s" %
              (node, info['system']['hostname']))

    if float(info['system']['uptime']) < LOW_UPTIME:
        warn("%s: Uptime is low: %ss" % (node, info['system']['uptime']))

    if float(info['system']['loadavg']) > HIGH_LOAD:
        warn("%s: 15 minute load average is %s" % (node, info['system']['loadavg']))

    for rpm in info['rpm']:
        if 'error' in rpm:
            if rpm['name'] not in rpm_versions:
                rpm_versions[rpm['name']] = {rpm['error']: [node]}
            else:
                versions = rpm_versions[rpm['name']]
                if rpm['error'] in versions:
                    versions[rpm['error']].append(node)
                else:
                    versions[rpm['error']] = [node]
        else:
            if rpm['name'] not in rpm_versions:
                rpm_versions[rpm['name']] = {rpm['version']: [node]}
            else:
                versions = rpm_versions[rpm['name']]
                if rpm['version'] in versions:
                    versions[rpm['version']].append(node)
                else:
                    versions[rpm['version']] = [node]
    for disk, use in info['disk']:
        use = int(use[:-1])
        if use > 90:
            warn("On %s, disk %s usage is %s%%", node, disk, use)

    for logfile, state in info['logrotate'].items():
        if not state:
            warn("%s: No log rotation configured for %s" % (node, logfile))

for cp in CORE_PACKAGES:
    if cp not in rpm_versions:
        error("Core package '%s' not installed on any node", cp)

for name, versions in rpm_versions.items():
    if len(versions) > 1:
        desc = ', '.join('%s (%s)' % (v, ', '.join(nodes)) for v, nodes in list(versions.items()))
        warn("Package %s: Versions differ! %s", name, desc)

    all_hosts = set(sum([hosts for hosts in list(versions.values())], []))
    for node in list(health_report.keys()):
        if len(all_hosts) > 0 and node not in all_hosts:
            warn("Package '%s' not installed on host '%s'" % (name, node))


def compare_system(systems):
    def check(value, msg):
        vals = set([system[value] for host, system in systems])
        if len(vals) > 1:
            info = ', '.join('%s: %s' % (h, system[value]) for h, system in systems)
            warn("%s: %s" % (msg, info))

    check('machine', 'Architecture differs')
    check('release', 'Kernel release differs')
    check('distname', 'Distribution differs')
    check('distver', 'Distribution version differs')
    # check('version', 'Kernel version differs')


def compare_files(systems):
    keys = set()
    for host, files in systems:
        keys.update(list(files.keys()))
    for filename in keys:
        vals = set([files.get(filename) for host, files in systems])
        if len(vals) > 1:
            info = ', '.join('%s: %s' % (h, files.get(filename)) for h, files in systems)
            warn("%s: %s" % ("Files differ", info))


compare_system((h, info['system']) for h, info in health_report.items())
compare_files((h, info['files']) for h, info in health_report.items())

if crm_script.output(2):
    report = crm_script.output(2)
    status = report.get('status')
    analysis = report.get('analysis')
    if status and not analysis:
        warn("Cluster report: %s" % (status))
    elif analysis:
        print("INFO: Cluster report:")
        print(analysis)
    else:
        warn("No cluster report generated")

if errors:
    for e in errors:
        print("ERROR:", e)
if warnings:
    for w in warnings:
        print("WARNING:", w)

if not errors and not warnings:
    print("No issues found.")

workdir = os.path.dirname(crm_script.__file__)
print("\nINFO: health-report in directory \"%s\"" % workdir)



#!/usr/bin/python3
import crm_script
try:
    uptime = open('/proc/uptime').read().split()[0]
    crm_script.exit_ok(uptime)
except Exception as e:
    crm_script.exit_fail("Couldn't open /proc/uptime: %s" % (e))



import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from crmsh import xmlutil, constants


class TestCrmMonXmlParser(unittest.TestCase):
    """
    Unitary tests for crmsh.xmlutil.CrmMonXmlParser
    """
    @mock.patch('crmsh.sh.cluster_shell')
    def setUp(self, mock_cluster_shell):
        """
        Test setUp.
        """
        data = '''
<data>
  <nodes>
    <node name="tbw-1" id="1084783148" online="true" standby="true" standby_onfail="false" maintenance="false" pending="false" unclean="false" shutdown="false" expected_up="true" is_dc="true" resources_running="3" type="member"/>
    <node name="tbw-2" id="1084783312" online="false" standby="false" standby_onfail="false" maintenance="false" pending="false" unclean="false" shutdown="false" expected_up="true" is_dc="false" resources_running="2" type="member"/>
  </nodes>
  <resources>
    <resource id="ocfs2-dlm" resource_agent="ocf:pacemaker:controld" role="Started" active="true" orphaned="false" blocked="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
      <node name="tbw-2" id="1084783312" cached="true"/>
    </resource>
    <resource id="ocfs2-clusterfs" resource_agent="ocf:heartbeat:Filesystem" role="Started" active="true" orphaned="false" blocked="false" managed="true" failed="false" failure_ignored="false" nodes_running_on="1">
      <node name="tbw-2" id="1084783312" cached="true"/>
    </resource>
  </resources>
</data>
        '''
        mock_cluster_shell().get_rc_stdout_stderr_without_input.return_value = (0, data, '')
        self.parser_inst = xmlutil.CrmMonXmlParser()

    def test_is_node_online(self):
        assert self.parser_inst.is_node_online("tbw-1") is True
        assert self.parser_inst.is_node_online("tbw-2") is False

    def test_get_node_list(self):
        assert self.parser_inst.get_node_list(standby=True) == ['tbw-1']
        assert self.parser_inst.get_node_list(online=False) == ['tbw-2']

    def test_is_resource_configured(self):
        assert self.parser_inst.is_resource_configured("test") is False
        assert self.parser_inst.is_resource_configured("ocf:heartbeat:Filesystem") is True

    def test_is_non_stonith_resource_running(self):
        assert self.parser_inst.is_non_stonith_resource_running() is True

    def test_is_resource_started(self):
        assert self.parser_inst.is_resource_started("test") is False
        assert self.parser_inst.is_resource_started("ocfs2-clusterfs") is True
        assert self.parser_inst.is_resource_started("ocf:pacemaker:controld") is True

    def test_get_resource_id_list_via_type(self):
        assert self.parser_inst.get_resource_id_list_via_type("test") == []
        assert self.parser_inst.get_resource_id_list_via_type("ocf:pacemaker:controld")[0] == "ocfs2-dlm"
